# -*- coding: utf-8 -*-
from operator import itemgetter
import ipaddress

# navigation chart as sankey diagram:
def navchartsankey(d, owndomain, omit):
    h = ''
    links = []    # {'source': 'google.com', 'target': '/team/view/ax', 'c': 1}
    nodes = []    # {'id': 'google.com', 'value':1}
    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        if len(k) < 8:   # not a day anymore
            break
        if 'user' in d['v0001']['days'][k] and 'navigation' in d['v0001']['days'][k]['user']:
            for sk,sv in d['v0001']['days'][k]['user']['navigation'].items():
                n = sk.split('(())') 
                if n[0] == n[1]:
                    continue
                if len(n[0]) == 0 or len(n[1]) == 0:
                    continue
                # allow only a-z and 0-9:
                if n[0] == '/':    # to avoid empty na
                    n[0] = owndomain
                if n[1] == '/':    # to avoid empty nb
                    n[1] = owndomain
                if any(oelm in n[0] for oelm in omit):  # do not show parts of url 
                    continue
                if any(oelm in n[1] for oelm in omit):  # do not show parts of url 
                    continue
                tmplink = {}
                tmplink['source'] = "".join(map(lambda char: char if char.isalnum()  else "", n[0]) ) # eliminate special chars
                tmplink['target'] = "".join(map(lambda char: char if char.isalnum()  else "", n[1]) ) # eliminate special chars
                tmplink['value'] = sv
                duplicate_found = False
                for li in links:
                    if (li['source'] == tmplink['source']
                            and li['target'] == tmplink['target']):
                        duplicate_found = True
                        li['value'] += sv
                        break
                    if tmplink['target'] == li['source']:  # prevent circular links
                        duplicate_found = True
                        break
                if duplicate_found == False:
                    links.append(tmplink)

                # add nodes if not already in array:
                n[0] = tmplink['source']
                n[1] = tmplink['target']
                for i in range(2):   # n[0] and n[1]
                    tmpnode = {}
                    tmpnode['name'] = n[i]
                    if n[i][0:1] != '/':
                        tmpnode['typ'] = 'root'
                    duplicate_found = False
                    tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[i]) ) # eliminate special chars
                    for no in nodes:
                        if (no['id'] == tmpnode['id']):
                            duplicate_found = True
                            break
                    
                    if duplicate_found == False:
                        nodes.append(tmpnode)
            
        days += 1
        if days > 31:
            break

    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # omit months or years:
        if len(k) <= 6:
            continue
        days += 1
        if 'externalFriendsHits' in d['v0001']['days'][k]['user']:
            for tk, tv in d['v0001']['days'][k]['user']['externalFriendsHits'].items():
                if is_valid_ip(tk) == True:  # to suppress ip; ip is not a domain anyway    
                    continue
                if '[' in tk:    # hack for IPv6 addresses
                    continue
                tmplink = {}
                tmplink['source'] =  "".join(map(lambda char: char if char.isalnum()  else "", tk) ) # eliminate special chars
                tmplink['value'] = 1
                for tdk,tdv in tv['target'].items():
                    if any(oelm in tdk for oelm in omit):  # don not show parts of url 
                        continue
                    tmplink['target'] =  "".join(map(lambda char: char if char.isalnum()  else "", tdk) ) # eliminate special chars
                    if tmplink['target'] == '/':    # to avoid empty nb
                        tmplink['target'] = owndomain
                    duplicate_found = False
                    for li in links:
                        if (li['source'] == tmplink['source']
                                and li['target'] == tmplink['target']):
                            duplicate_found = True
                            li['value'] += tdv
                            break
                        if tmplink['target'] == li['source']:  # prevent circular links
                            duplicate_found = True
                            break

                    if duplicate_found == False:
                        links.append(tmplink)

                    n[0] = tmplink['source'] # tk
                    n[1] = tmplink['target'] #tdk
                    for i in range(2):   # n[0] and n[1]
                        tmpnode = {}
                        tmpnode['name'] = n[i]
                        tmpnode['typ'] = 'root'
                        duplicate_found = False
                        tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[i]) ) # eliminate special chars
                        for no in nodes:
                            if (no['id'] == tmpnode['id']):
                                duplicate_found = True
                                break
                    
                        if duplicate_found == False:
                            nodes.append(tmpnode)
                days += 1
                if days > 31:
                    break

                    
 
    # d3js horizontal bubble char in case results are available:
    h += '<script src="https://cdn.jsdelivr.net/npm/d3-sankey@0.12.3/dist/d3-sankey.min.js"></script>'
    h += "\n\n"
    h += '<div class="col-md-12 col-lg-12 col-xxl-12">'
    h += '<div class="card mt-2"><div class="card-body">'
    h += '<h3 class="card-title">Navigation Chart</h3>'
    h += '<p class="card-text">User Navigation Chart for ' + owndomain + ':</p>'
    h += '<div id="navchart-container20250128"><svg id="svgchart20250128" width="600" height="400"></svg></div>'
    h += '<script type="module">' + "\n"
    h += 'const nodes = ' + str(nodes) + ';' + "\n"
    h += 'const links = ' + str(links) + ';' + "\n"
    h += 'const rect = document.getElementById("navchart-container20250128").getBoundingClientRect();'
    h += 'const margin = { top: 1, right: 20, bottom: 5, left: 1 };'
    h += 'const width = Math.round(rect.width) - margin.left - margin.right;'
    h += 'const height = width;'  + "\n"   # make height at least as width

    #  Convert the source and target to internal node indices
    h += 'const LS = d3.map(links, (link) => link.source);'
    h += 'const LT = d3.map(links, (link) => link.target);'
    h += 'const LV = d3.map(links, (link) => link.value);'

    h += 'const N = d3.map(nodes, (node) => node.id);'
    h += 'const G = null;'   # No specific node groups in this case

    #// Create a sankey layout
    h += 'const sankey = d3.sankey()'
    h += '.nodeId(({ index: i }) => N[i])'
    h += '.nodeWidth(10)'
    h += '.nodePadding(10)'
    h += '.nodeAlign(d3.sankeyJustify)'
    h += '.extent(['
    h += '[margin.left, margin.top],'
    h += '[width - margin.right, height - margin.bottom]'
    h += ']);'  + "\n"

    h += 'const { nodes: computedNodes, links: computedLinks } = sankey({'
    h += 'nodes: nodes,'
    h += 'links: links'
    h += '});'  + "\n"

    #// Create the SVG container
    h += 'const svg = d3.select("#svgchart20250128")'
    h += '.attr("width", width)'
    h += '.attr("height", height)'
    h += '.attr("viewBox", [0,0 , width, height])'
    h += '.attr("style", "max-width: 100%; height: auto;  font: 10px sans-serif; height: intrinsic;");'  + "\n"

    # // Add node elements (rectangles)
    h += 'const node = svg'
    h += '.append("g")'
    h += '.attr("stroke", "currentColor")'
    h += '.selectAll("rect")'
    h += '.data(computedNodes)'
    h += '.join("rect")'
    h += '.attr("x", (d) => d.x0)'
    h += '.attr("y", (d) => d.y0)'
    h += '.attr("width", (d) => d.x1 - d.x0)'
    h += '.attr("height", (d) => d.y1 - d.y0)'
    h += '.attr("fill", "#1f77b4");'  + "\n"

    #// Add labels to the nodes
    h += 'svg'
    h += '.append("g")'
    h += '.attr("font-family", "sans-serif")'
    h += '.attr("font-size", 10)'
    h += '.selectAll("text")'
    h += '.data(computedNodes)'
    h += '.join("text")'
    h += '.attr("x", (d) => (d.x0 < width / 2 ? d.x1 + 6 : d.x0 - 6))'
    h += '.attr("y", (d) => (d.y1 + d.y0) / 2)'
    h += '.attr("dy", "0.35em")'
    h += '.attr("text-anchor", (d) => (d.x0 < width / 2 ? "start" : "end"))'
    h += '.text((d) => d.id);'  + "\n"

    #// Add link elements (paths)
    h += 'const link = svg'
    h += '.append("g")'
    h += '.attr("fill", "none")'
    h += '.attr("stroke-opacity", 0.5)'
    h += '.selectAll("g")'
    h += '.data(computedLinks)'
    h += '.join("g");'  + "\n"

    h += 'link'
    h += '.append("path")'
    h += '.attr("d", d3.sankeyLinkHorizontal())'
    h += '.attr("stroke", "#ccc")'
    h += '.attr("stroke-width", (d) => Math.max(1, d.width));'  + "\n"

    #// Add titles to the links
    h += 'link'
    h += '.append("title")'
    h += '.text((d) => `${d.source.id} → ${d.target.id}\n${d.value}`);'  + "\n"

    h += '</script>'
    h += '</div></div></div>'
    return h

def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False



import json
from copy import deepcopy

def migv0001(
    statfile
    ):
    try:
        f = open(statfile) 
        d = json.load(f) 
    except FileNotFoundError:   # first call: file does not exists
        print("-s statistic file not found, it will be automatically created")
        return
    except json.JSONDecodeError:
        print("-s json file is not valid")
        return

    # return if already migrated:
    if 'v0001' in d:
        return

    print("migrateV0001: Statfile to be migrated to v0001: " + statfile)

    # write Save
    tstat = statfile.replace('.json','v0001before.json')
    with open(tstat, "w") as sf:
       json.dump(d,sf)

    d['v0001'] = {}
    d['v0001']['days'] = {}
    for x in d['days']:
        d['v0001']['days'][x] = {}
        d['v0001']['days'][x]['user'] = {}
        d['v0001']['days'][x]['user']['visits'] = 0
        if 'visits' in d['days'][x]:
            d['v0001']['days'][x]['user']['visits'] = int(d['days'][x]['visits'])
            
        d['v0001']['days'][x]['user']['bytesSent'] = 0
        if 'bytes_sent' in d['days'][x]:
            d['v0001']['days'][x]['user']['bytesSent'] = int(d['days'][x]['bytes_sent'])
            
        d['v0001']['days'][x]['user']['countryHits'] = d['days'][x]['countries']
        
        d['v0001']['days'][x]['user']['deviceHits'] = {}
        d['v0001']['days'][x]['user']['deviceHits']['desktop'] = d['days'][x]['device_hits']['desktop']
        d['v0001']['days'][x]['user']['deviceHits']['mobile'] = d['days'][x]['device_hits']['mobile']
        d['v0001']['days'][x]['user']['deviceHits']['tablet'] = d['days'][x]['device_hits']['tablet']
        
        d['v0001']['days'][x]['user']['externalFriendsHits'] = {}
        if 'friends' in d['days'][x]:
            d['v0001']['days'][x]['user']['externalFriendsHits'] = d['days'][x]['friends']
        d['v0001']['days'][x]['user']['topUrl'] = {}
        if 'topurl' in d['days'][x]:
            d['v0001']['days'][x]['user']['topUrl'] = d['days'][x]['topurl']
            
        d['v0001']['days'][x]['user']['serverResponseCode'] = {}
        if 'serverResponseCode' in d['days'][x]:
            d['v0001']['days'][x]['user']['serverResponseCode'] = d['days'][x]['serverResponseCode']

        d['v0001']['days'][x]['robot'] = {}
        d['v0001']['days'][x]['robot']['bytesSent'] = 0
        d['v0001']['days'][x]['robot']['robotHits'] = 0
        if 'bots' in d['days'][x]['device_hits']:
            d['v0001']['days'][x]['robot']['robotHits'] += d['days'][x]['device_hits']['bots'] 
        if 'others' in d['days'][x]['device_hits']:
            d['v0001']['days'][x]['robot']['robotHits'] += d['days'][x]['device_hits']['others'] 
            
        d['v0001']['days'][x]['quality'] = {}
        if 'quality' in d['days'][x]:
            d['v0001']['days'][x]['quality'] = d['days'][x]['quality'] 

    # write updated statistic file:
    with open(statfile, "w") as sf:
       json.dump(d,sf)

    return

# delete v0000 version of data
def delv0000(
    statfile
    ):
    try:
        f = open(statfile) 
        d = json.load(f) 
    except FileNotFoundError:   # first call: file does not exists
        print("-s statistic file not found")
        return
    except json.JSONDecodeError:
        print("-s json file is not valid")
        return

    # return if not migrated:
    if 'v0001' not in d:
        return

    # return if already migrated:
    if 'days' not in d:
        return

    #dtmp = deepcopy(d)
    # delete old data only if it exists in v0001:
    #for x in dtmp['days']:
    #    if x in d['v0001']['days']:
    #      del d['days'][x]
    del d['days']

    # write updated statistic file:
    with open(statfile, "w") as sf:
       json.dump(d,sf)

    return


        




# -*- coding: utf-8 -*-
from operator import itemgetter
import ipaddress

# navigation chart:
def navchart(d, owndomain, omit):
    h = ''
    links = []    # [{'source': 'google.com', 'target': '/team/view/ax', 'cnt': 2, 'source': 'bing.com', 'target': '/team/viex/comp', 'cnt': 5}]
    nodes = []    # [{'id': 'google.com', 'root': 'y', 'cnt_in': 1, 'cnt_out': 2}, {'id': '/team/view/ax/', 'cnt_in': 2, 'cnt_out': 1}]
    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        if len(k) < 8:   # not a day anymore
            break
        if 'user' in d['v0001']['days'][k] and 'navigation' in d['v0001']['days'][k]['user']:
            for sk,sv in d['v0001']['days'][k]['user']['navigation'].items():
                n = sk.split('(())') 
                if n[0] == n[1]:
                    continue
                if len(n[0]) == 0 or len(n[1]) == 0:
                    continue
                # allow only a-z and 0-9:
                if n[0] == '/':    # to avoid empty na
                    n[0] = owndomain
                if n[1] == '/':    # to avoid empty nb
                    n[1] = owndomain
                if any(oelm in n[0] for oelm in omit):  # do not show parts of url 
                    continue
                if any(oelm in n[1] for oelm in omit):  # do not show parts of url 
                    continue
                tmplink = {}
                tmplink['source'] = "".join(map(lambda char: char if char.isalnum()  else "", n[0]) ) # eliminate special chars
                tmplink['target'] = "".join(map(lambda char: char if char.isalnum()  else "", n[1]) ) # eliminate special chars
                tmplink['cnt'] = sv
                duplicate_found = False
                for li in links:
                    if (li['source'] == tmplink['source']
                            and li['target'] == tmplink['target']):
                        duplicate_found = True
                        li['cnt'] += sv
                        break
                if duplicate_found == False:
                    links.append(tmplink)

                for i in range(2):   # n[0] and n[1]
                    tmpnode = {}
                    tmpnode['name'] = n[i]
                    tmpnode['cnt_in'] = 0
                    tmpnode['cnt_out'] = 0
                    if ('.com' in n[i] or
                           '.ru' in n[i]):
                        tmpnode['root'] = 'y'
                    duplicate_found = False
                    tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[i]) ) # eliminate special chars
                    if i == 0:
                        tmpnode['cnt_in'] = tmplink['cnt']
                    else:
                        tmpnode['cnt_out'] = tmplink['cnt']
                    
                    for no in nodes:
                        if (no['id'] == tmpnode['id']):
                            duplicate_found = True
                            if i == 0:
                                no['cnt_in'] += tmpnode['cnt_in']
                            else:
                                no['cnt_out'] += tmpnode['cnt_out']
                            break
                    
                    if duplicate_found == False:
                        nodes.append(tmpnode)
            
        days += 1
        if days > 31:
            break

    days = 0
    for k, v in sorted(d['v0001']['days'].items(), key=itemgetter(0), reverse=True):
        # omit months or years:
        if len(k) <= 6:
            continue
        days += 1
        if 'externalFriendsHits' in d['v0001']['days'][k]['user']:
            for tk, tv in d['v0001']['days'][k]['user']['externalFriendsHits'].items():
                if is_valid_ip(tk) == True:  # to suppress ip; ip is not a domain anyway    
                    continue
                if '[' in tk:    # hack for IPv6 addresses
                    continue
                tmplink = {}
                tmplink['source'] =  "".join(map(lambda char: char if char.isalnum()  else "", tk) ) # eliminate special chars
                tmplink['cnt'] = 0
                for tdk,tdv in tv['target'].items():
                    if any(oelm in tdk for oelm in omit):  # don not show parts of url 
                        continue
                    tmplink['target'] =  "".join(map(lambda char: char if char.isalnum()  else "", tdk) ) # eliminate special chars
                    duplicate_found = False
                    for li in links:
                        if (li['source'] == tmplink['source']
                                and li['target'] == tmplink['target']):
                            duplicate_found = True
                            li['cnt'] += tdv
                            break
                    if duplicate_found == False:
                        links.append(tmplink)

                    n[0] = tk
                    n[1] = tdk
                    for i in range(2):   # n[0] and n[1]
                        tmpnode = {}
                        tmpnode['name'] = n[i]
                        tmpnode['cnt_in'] = 0
                        tmpnode['cnt_out'] = 0
                        tmpnode['root'] = 'y'
                        duplicate_found = False
                        tmpnode['id'] = "".join(map(lambda char: char if char.isalnum()  else "", n[i]) ) # eliminate special chars
                        for no in nodes:
                            if (no['id'] == tmpnode['id']):
                                duplicate_found = True
                                if i == 0:
                                    no['cnt_in'] += tmpnode['cnt_in']
                                else:
                                    no['cnt_out'] += tmpnode['cnt_out']
                                break
                    
                        if duplicate_found == False:
                            nodes.append(tmpnode)
                days += 1
                if days > 31:
                    break


    # start with the top 10 root nodes sorted by  'cnt_out' 
    root_nodes = [item for item in nodes if item.get('root') == 'y']
    top_nodes = sorted(root_nodes, key=lambda x: x['cnt_out'], reverse=True)
    print("top_nodes: " + str(top_nodes) )

    chart_links = []
    chart_nodes = []
    for n in top_nodes:
        for li in links:
            if n['id'] == li['source']:
                chart_links.append(li)
                # check source_node:
                cn_found = False
                for cn in chart_nodes:
                    if li['source'] == cn['id']:
                        cn_found = True
                if cn_found == False:
                    tmpnode = {}
                    tmpnode['id'] = li['source']
                    tmpnode['name'] = li['source']
                    for nn in nodes:
                        if nn['id'] == li['source']:
                            tmpnode['name'] = nn['name']
                            break
                    if n['root'] == 'y':
                        tmpnode['root'] = 'y'
                    chart_nodes.append(tmpnode)
                # check target_node:
                cn_found = False
                for cn in chart_nodes:
                    if li['target'] == cn['id']:
                        cn_found = True
                if cn_found == False:
                    tmpnode = {}
                    tmpnode['name'] = li['target']
                    tmpnode['id'] = li['target']
                    for nn in nodes:
                        if nn['id'] == li['target']:
                            tmpnode['name'] = nn['name']
                            break
                    chart_nodes.append(tmpnode)
 
    # d3js horizontal bubble char in case results are available:
    h += "\n\n"
    h += '<div class="col-md-12 col-lg-12 col-xxl-12">'
    h += '<div class="card mt-2"><div class="card-body">'
    h += '<h3 class="card-title">Navigation Chart</h3>'
    h += '<p class="card-text">User Navigation Chart for ' + owndomain + ':</p>'
    h += '<div id="navchart-container"><svg id="svgchart" width="600" height="400"></svg></div>'
    h += '<script type="module">' + "\n"
    h += 'const nodes = ' + str(chart_nodes) + ';' + "\n"
    h += 'const links = ' + str(chart_links) + ';' + "\n"
    h += 'const rect = document.getElementById("navchart-container").getBoundingClientRect();'
    h += 'const margin = { top: 20, right: 20, bottom: 40, left: 100 };'
    h += 'const width = Math.round(rect.width) - margin.left - margin.right;'
    #h += 'const height = 400 - margin.top - margin.bottom;'  + "\n"
    h += 'const height = width;'  + "\n"   # make height at least as width

    # 2. Set up the SVG container and set dimensions
    h += 'const svg = d3.select("#svgchart").attr("width", width).attr("height", height);'
      
    # 3. Set up the D3 force simulation
    h += "const validLinks = links.filter(link => link.source && link.target);"
    h += "const simulation = d3.forceSimulation(nodes)"
    h += ".force('link', d3.forceLink(validLinks).id(d => d.id).distance(100))"
    h += ".force('charge', d3.forceManyBody().strength(-300))"
    h += ".force('center', d3.forceCenter(width / 2, height / 2));"  + "\n"

    # 4. Create links (edges)
    h += "const link = svg.append('g')"
    h += ".selectAll('.link')"
    h += ".data(links)"
    h += ".enter().append('line')"
    h += ".style('stroke','#999')"
    h += ".style('stroke-opacity','0.6');" + "\n"
    #h += ".attr('class', 'link');"  + "\n"

    # 5. Create nodes
    h += "const node = svg.append('g')"
    h += ".selectAll('.node')"
    h += ".data(nodes)"
    h += ".enter().append('circle')"
    h += ".attr('class', 'node')"
    h += ".attr('r', 10)"
    h += ".style('fill','steelblue')"
    h += ".style('stroke','white')"
    h += ".style('stroke-width','1.5px')"
    h += ".call(d3.drag()"
    h += ".on('start', dragStarted)"
    h += ".on('drag', dragged)"
    h += ".on('end', dragEnded));" + "\n"

    # 6. Add labels to the nodes
    h += "const labels = svg.append('g')"
    h += ".selectAll('.label')"
    h += ".data(nodes)"
    h += ".enter().append('text')"
    h += ".style('font-family','Arial, sans-serif')"
    h += ".style('font-size','10px')"
    h += ".style('pointer-events','none')"
    h += ".attr('dx', 25)"
    h += ".attr('dy', '.35em')"
    h += ".text(d => {"
    h += "  let lbl;"
    h += "  if (d.name.length > 15) {"
    h += "    lbl = d.name.substring(0,10) + '...' + d.name.slice(-3);"
    h += "  } else {"
    h += "    lbl = d.name;"
    h += "  }"
    h += "  return lbl; }"
    h += ");"  + "\n"

    # 7. Place root nodes along the left side of the screen and freeze their positions
    h += "const rootNodes = nodes.filter(n => n.root === 'y');"
    h += "rootNodes.forEach((node, index) => {"
    h += "node.x = 100;" # Set all root nodes on the left (x = 100)
    h += "node.y = 100 + index * 100;"   # Distribute vertically along a line
    h += "node.fx = node.x;"  # Freeze position
    h += "node.fy = node.y;"  # Freeze position
    h += "d3.selectAll('circle.node').filter(d => d && d.id === node.id).style('fill', 'red');"  # color red
    h += "});"  + "\n"

    # 8. Define the tick function to update positions
    h += "simulation.on('tick', function() {"
    h += "link"
    h += ".attr('x1', d => d.source.x)"
    h += ".attr('y1', d => d.source.y)"
    h += ".attr('x2', d => d.target.x)"
    h += ".attr('y2', d => d.target.y);"  + "\n"

    # Update node positions
    h += "node.attr('cx', d => d.x).attr('cy', d => d.y);"  + "\n"

    # Update label positions
    h += "labels.attr('x', d => d.x).attr('y', d => d.y);"
    h += "});"  + "\n"

    # 9. Drag functions to allow node movement
    h += "function dragStarted(event) {"
    h += "if (!event.active) simulation.alphaTarget(0.3).restart();"
    h += "event.subject.fx = event.subject.x;"
    h += "event.subject.fy = event.subject.y;"
    h += "}"  + "\n"

    h += "function dragged(event) {"
    h += "event.subject.fx = event.x;"
    h += "event.subject.fy = event.y;"
    h += "}"  + "\n"

    h += "function dragEnded(event) {"
    h += "if (!event.active) simulation.alphaTarget(0);"
    h += "event.subject.fx = null;"
    h += "event.subject.fy = null;"
    h += "}"  + "\n"
    h += "</script>"

    h += '</div></div></div>'
    return h

def is_valid_ip(address):
    try: 
        x = ipaddress.ip_address(address)
        return True
    except:
        return False
